// written by Zheng Gengbin
// optimization
// 1. local index => global index
// 2. row/column major

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
#endif

#include <stdio.h>
#include "compilerApi.h"

#include "anls_att.h"

#include "detect_comm.h"

//extern "C" int strcasecmp(...);

//---------------------------------------

// a(i)=i , i is translate to a func call pcrc_glob , which is time comsuming ,
// now , we  change it to
//if (pcrc_on (pcrc_group (pcrc_sdd0))) then
//          pcrc_step_i0 = pcrc_glob (pcrc_irg0,pcrc_lil_i)
//         pcrc_step_i = pcrc_glob (pcrc_irg0,pcrc_lis_i+pcrc_lil_i)-pcrc_&
//     &glob (pcrc_irg0,pcrc_lil_i)
//         do  i=0,(pcrc_liu_i-pcrc_lil_i)/pcrc_lis_i,1
//            a(pcrc_sdx0) = pcrc_step_i0
//            pcrc_sdx0 = pcrc_sdx0+pcrc_u00
//            pcrc_step_i0 = pcrc_step_i0+pcrc_step_i
//         enddo
//       endif

#define MAX_INDEX  10

#define INDEX_EXP  1000

void get_scalar_exp(SgStatement *scope,SgExpression *exp)
{
	if (exp)
	switch(exp->variant())
	{
	case FUNC_CALL:
		if ( strcasecmp(exp->symbol()->identifier(),"pcrc_glob") == 0 )
		{
//			SgExpression *index=exp->lhs()->rhs()->lhs()->lhs()->rhs();
			  // record pcrc_glob()
			scope->addAttribute(INDEX_EXP,exp,sizeof(exp) );
		}
		break;
	default:
		get_scalar_exp(scope,exp->lhs());
		get_scalar_exp(scope,exp->rhs());
		break;
	}
}

void get_scalar_stmt(SgStatement *scope,SgStatement *s)
{
   switch (s->variant() )
   {
   case FOR_NODE: {
        for (int i=0; i<s->numberOfChildrenList1(); i++)
		get_scalar_stmt(scope,s->childList1(i));
	break;
	}
   case LOGIF_STMT: 
   case LOGIF_NODE: {
	SgIfStmt *ifstmt = (SgIfStmt*)(s);
	get_scalar_exp(scope,ifstmt->conditional());
        for (int i=0; i<s->numberOfChildrenList1(); i++)
		get_scalar_stmt(scope,s->childList1(i));
	break;
  	}
   case ASSIGN_STAT: {
	get_scalar_exp(scope,s->expr(1));
	break;
	}
   case ENDDO_STMT:
   case CONTROL_END: {
	break;
	}
   default: {
//	printf("abnormal:%d!\n",s->variant());
	break;
	}
   }
}

void add_init_value(SgStatement *s,SgStatement *loop,SgSymbol *scalarsym,SgSymbol *irg)
{
	SgSymbol * step0 = s->getSymbolInCurrentScope(mixstring, "step", scalarsym->identifier(),0);

	SgSymbol * lil = s->getSymbolInCurrentScope(mixstring, "lil", scalarsym->identifier());
	SgSymbol * lis = s->getSymbolInCurrentScope(mixstring, "lis", scalarsym->identifier());

	SgSymbol * step = s->getSymbolInCurrentScope(mixstring, "step", scalarsym->identifier());

	{
	SgExpression *lilexp = new SgVarRefExp(lil);
	SgExpression *globexp = make_glob(s,irg,lilexp);
	SgExpression *step0exp = new SgVarRefExp(step0);

	// add pcrc_step_i0 = pcrc_glob(pcrc_irg0 , pcrc_lil_i)
	SgAssignStmt *ass = new SgAssignStmt(*step0exp , *globexp);
	loop->insertStmtBefore(*ass, *loop->controlParent());
	}

	{
	SgExpression *stepexp = new SgVarRefExp(step);
	SgExpression *lilexp = new SgVarRefExp(lil);
	SgExpression *lisexp = new SgVarRefExp(lis);
	SgExpression *exp = &(*lisexp  + *lilexp );
	SgExpression *globexp1 = make_glob(s,irg,exp);
	SgExpression *globexp2 = make_glob(s,irg,lilexp);
	exp = &( *globexp1 - *globexp2 );
	SgAssignStmt *ass = new SgAssignStmt(*stepexp , *exp);
	SgStatement *do_loop = loop;
	while ( do_loop->variant() == FOR_NODE  && do_loop->controlParent() &&
		do_loop->controlParent()->variant() == FOR_NODE )
	{
		do_loop=do_loop->controlParent();
	}
	do_loop->insertStmtBefore(*ass, *do_loop->controlParent());
	}

	// add increase
	{
	SgExpression *step0exp = new SgVarRefExp(step0);
	SgExpression *stepexp = new SgVarRefExp(step);
	SgExpression *exp = &(*step0exp + *stepexp );
	SgAssignStmt *ass = new SgAssignStmt(*step0exp->copyPtr(), *exp);
	SgStatement *laststmt = loop->childList1(loop->numberOfChildrenList1()-3);	// enddo stmt , control_end
	ASSERT(laststmt);
	laststmt->insertStmtAfter(*ass, *laststmt->controlParent());
	}
}

void optimize_index(SgStatement *assign,Forall_header_def *fhdef)
{
  int i,j;

  SgIfStmt *ifstmt=NULL;
  get_scalar_exp(assign,assign->expr(1));
  if (assign->controlParent()->variant() == LOGIF_NODE)
  {
	ifstmt = isSgIfStmt(assign->controlParent());
	get_scalar_exp(assign,ifstmt->conditional());
  }

  for (i=0; i<fhdef->getindexnumber(); i++) {
     int add=0;
     for (j=0; j<assign->numberOfAttributes(INDEX_EXP); j++) {
	SgExpression *exp = (SgExpression *)(assign->attributeValue(j,INDEX_EXP));
	SgExpression *index=exp->lhs()->rhs()->lhs()->lhs()->rhs();
	SgSymbol *irg=exp->lhs()->lhs()->symbol();
	SgSymbol *scalarsym = index->symbol();
//printf("scalarsym:%s,%s\n",scalarsym->identifier(),fhdef->getindexsymb(i)->identifier());

	if(strcasecmp(fhdef->getindexsymb(i)->identifier(),scalarsym->identifier()) == 0) 
        {
		add=1;
		// find the loop of index
		SgStatement *do_loop = ifstmt?ifstmt->controlParent():assign->controlParent();
		while ( do_loop->variant() == FOR_NODE )
		{
			if ( strcasecmp( ((SgForStmt *)do_loop)->doName()->identifier() ,scalarsym->identifier()) == 0 )
			break;
			do_loop=do_loop->controlParent();
		}
  		// add init value of pcrc_step_i0 and pcrc_step_i
if (!silentFlag)
printf("optimize local=>global:%s\n",scalarsym->identifier());
		ASSERT(do_loop);
		add_init_value(assign,do_loop,scalarsym,irg);
        }
        if (add == 1) break;
     }
  }

    // change  pcrc_glob => pcrc_step_i0
  for (i=0; i<assign->numberOfAttributes(INDEX_EXP); i++) {
	SgExpression *exp = (SgExpression *)(assign->attributeValue(i,INDEX_EXP));
	SgExpression *index=exp->lhs()->rhs()->lhs()->lhs()->rhs();
	SgSymbol *scalarsym = index->symbol();
	SgSymbol * step0 = assign->getSymbolInCurrentScope(mixstring, "step", scalarsym->identifier(),0);
	SgExpression *step0exp = new SgVarRefExp(step0);
	exp->replace(*step0exp);
  }

}

// main function
// scope is the innest do loop
void optimize_glob_index(SgStatement *scope,Forall_header_def *fhdef)
{
  int i,j;

  scope->deleteAllAttribute(INDEX_EXP);
  get_scalar_stmt(scope,scope);

  for (i=0; i<fhdef->getindexnumber(); i++) {
     int add=0;
     for (j=0; j<scope->numberOfAttributes(INDEX_EXP); j++) {
	SgExpression *exp = (SgExpression *)(scope->attributeValue(j,INDEX_EXP));
	SgExpression *index = exp->lhs()->rhs()->lhs()->lhs()->rhs();
	SgSymbol *irg = exp->lhs()->lhs()->symbol();
	SgSymbol *scalarsym = index->symbol();

	if(strcasecmp(fhdef->getindexsymb(i)->identifier(),scalarsym->identifier()) == 0) 
        {
		add=1;
		  // find the loop of index
		SgStatement *do_loop = scope;
		while ( do_loop->variant() == FOR_NODE )
		{
			if ( strcasecmp( ((SgForStmt *)do_loop)->doName()->identifier() ,scalarsym->identifier()) == 0 )
			break;
			do_loop=do_loop->controlParent();
		}
  		// add init value of pcrc_step_i0 and pcrc_step_i
if (!silentFlag)
printf("optimize local=>global:%s\n",scalarsym->identifier());
		ASSERT(do_loop);
		add_init_value(scope,do_loop,scalarsym,irg);
        }
        if (add == 1) break;
     }
  }

    // change  pcrc_glob => pcrc_step_i0
  for (i=0; i<scope->numberOfAttributes(INDEX_EXP); i++) {
	SgExpression *exp = (SgExpression *)(scope->attributeValue(i,INDEX_EXP));
	SgExpression *index=exp->lhs()->rhs()->lhs()->lhs()->rhs();
	SgSymbol *scalarsym = index->symbol();
	SgSymbol * step0 = scope->getSymbolInCurrentScope(mixstring, "step", scalarsym->identifier(),0);
	SgExpression *step0exp = new SgVarRefExp(step0);
	exp->replace(*step0exp);
  }

}


/***************************************************************************
	compare_distribute_pattern( actu , dummy )

	compare if the distribution pattern of actual argument and dummy
are the same .
	actu: have ALIGN , and DISTRIBUTE
	dummy: only have DISTRIBUTE
	
***************************************************************************/

int compare_distribute_pattern(SgSymbol *actu , SgSymbol *dummy)
{
	int lb,ub;

if (!silentFlag)
printf("compare distribute pattern:%s,%s\n",actu->identifier(),dummy->identifier() );

	if (actu->numberOfAttributes(Align_Def) == 0) return 0;
	Alignment_def *adef=(Alignment_def *)actu->attributeValue(0,Align_Def);
  	SgSymbol * templatesym = adef->gettosymbol();
	ASSERT (templatesym->numberOfAttributes(Temp_Def) != 0) ;
  	Template_def * tdef = (Template_def *) ((templatesym->getAttribute(0, Temp_Def))->getAttributeData());
	ASSERT (templatesym->numberOfAttributes(Dist_Def) != 0) ;
  	Distribute_def * ddef1 = (Distribute_def *) templatesym->getAttribute(0, Dist_Def)->getAttributeData();

	if (dummy->numberOfAttributes(Dist_Def) == 0) return 0;
	Distribute_def *ddef2=(Distribute_def *)dummy->attributeValue(0,Dist_Def);

	// align with one by one
	int noop = 0;
	for (int i=0; i< adef->getDim() ; i++) {
		int j=adef->getAlignPair(i);
		// collapse
		if (j == -1 ) {
			if (ddef2->distCode(i) != 0 ) return 0;
			continue;
		}
		else {
			if ( ddef1->distCode(noop) != ddef2->distCode(i) ) 
				return 0;
			ASSERT(adef->getStride(i));
			int a=adef->getStride(i)->valueInteger();
  			int b=adef->getOffset(i)->valueInteger();
			if ( a!= 1 ) return 0;
			if ( b!= 0 ) return 0;
		}
		  // compare the size
		if (dummy->isAssumedShapeSymbol())
		{
			// next
		}
		else if (dummy->isAssumedSizeSymbol())
		{
			return 0;
		}
		else
		{
		int t_lb=tdef->getStart(j)->valueInteger();
  		int t_ub=tdef->getEnd(j)->valueInteger();
		SgExpression *lbexp  = get_lbound(dummy,i);
		SgExpression *ubexp  = get_ubound(dummy,i);
		if (lbexp->isInteger()) 
			lb = lbexp->valueInteger();
		else
			return 0;
		if (ubexp->isInteger()) 
			ub = ubexp->valueInteger();
		else
			return 0;
		if (t_lb != lb) return 0;
		if (t_ub != ub) return 0;
		}
		noop++;
	}

if (!silentFlag)
printf("compare same.\n");
	return 1;
}


/***************************************************************************
	compare_data_mapping_pattern( lhs , rhs )

	compare if the distribution pattern of two array are the same .
        now allow normal array and dummy which has only distribute directive.
***************************************************************************/

int compare_data_mapping_pattern(forall_reference *fref1 , forall_reference *fref2,Forall_header_def *fhdef)
{
	int lb,ub;

	SgArrayRefExp *arrayexp1 = (SgArrayRefExp*)(fref1->getexpression());
	SgArrayRefExp *arrayexp2 = (SgArrayRefExp*)(fref2->getexpression());
	SgSymbol *a = arrayexp1->symbol();
	SgSymbol *b = arrayexp2->symbol();
if (!silentFlag)
printf("Compare data mapping pattern:%s,%s ...\n",a->identifier(),b->identifier() );

	Distribute_def *ddef1=NULL,*ddef2=NULL;
        Alignment_def *adef1=NULL,*adef2=NULL;
        Template_def * tdef1=NULL,*tdef2=NULL;

	if (a->numberOfAttributes(Align_Def) != 0) {
	   adef1=(Alignment_def *)a->attributeValue(0,Align_Def);
  	   SgSymbol * templatesym1 = adef1->gettosymbol();
	   // if align to a dummy , no Temp_Def.
	   if (templatesym1->numberOfAttributes(Temp_Def) != 0)
  	      tdef1 = (Template_def *) ((templatesym1->getAttribute(0, Temp_Def))->getAttributeData());
	   ASSERT (templatesym1->numberOfAttributes(Dist_Def) != 0) ;
  	   ddef1 = (Distribute_def *) templatesym1->getAttribute(0, Dist_Def)->getAttributeData();
        }
      	else {
	   if (a->numberOfAttributes(Dist_Def) == 0) return 0;      // INHERIT
  	   ddef1 = (Distribute_def *) a->getAttribute(0, Dist_Def)->getAttributeData();
	}

	if (b->numberOfAttributes(Align_Def) != 0) {
	   adef2=(Alignment_def *)b->attributeValue(0,Align_Def);
  	   SgSymbol * templatesym2 = adef2->gettosymbol();
	   if (templatesym2->numberOfAttributes(Temp_Def) != 0)
  	      tdef2 = (Template_def *) ((templatesym2->getAttribute(0, Temp_Def))->getAttributeData());
	   ASSERT (templatesym2->numberOfAttributes(Dist_Def) != 0) ;
  	   ddef2 = (Distribute_def *) templatesym2->getAttribute(0, Dist_Def)->getAttributeData();
	}
	else {
	   if (b->numberOfAttributes(Dist_Def) == 0) return 0;      // INHERIT
  	   ddef2 = (Distribute_def *) b->attributeValue(0, Dist_Def);
        }

	// process is not same
	if (ddef1->gettosymbol() != ddef2->gettosymbol())
	{
if (!silentFlag)
printf("Processors are different.\n");
		return 0;
	}

	for (int sub1=0; sub1<arrayexp1->numberOfSubscripts(); sub1++) {
		int index = fref1->indexNumberOnDim(sub1);
		if (index == -1) continue;
		int sub2 = fref2->dimNumberOnIndex(index);
		//ASSERT (sub2!=-1);
		if (sub2 == -1) continue;

		// compare the align offset & stride
		SgExpression *str1 = adef1?adef1->getStride(sub1):new SgValueExp(1);
		SgExpression *str2 = adef2?adef2->getStride(sub2):new SgValueExp(1);
		SgExpression *off1 = adef1?adef1->getOffset(sub1):new SgValueExp(0);
		SgExpression *off2 = adef2?adef2->getOffset(sub2):new SgValueExp(0);
		// str == NULL, means this dim is collapsed.
		if (str1 == NULL && str2 == NULL)
		{
			// pass
		}
		else if (str1 == NULL || str1 == NULL)
		{
if (!silentFlag)
printf("%s's Align dummy in dim %d is different with %s's dim %d.\n",a->identifier(),sub1+1,b->identifier(),sub2+1);
			return 0; 
		}
		else if (*str1 != *str2 || *off1 != *off2)
		{
if (!silentFlag)
printf("%s's Align offset and stride in dim %d is different with %s's dim %d.\n",a->identifier(),sub1+1,b->identifier(),sub2+1);
			return 0; 
		}
		
		int align1 = adef1?adef1->getAlignPair(sub1):sub1;
		int align2 = adef2?adef2->getAlignPair(sub2):sub2;

		SgExpression *lb1 = tdef1?tdef1->getStart(align1):get_lbound(a,sub1);
		SgExpression *lb2 = tdef2?tdef2->getStart(align2):get_lbound(b,sub2);
		SgExpression *ub1 = tdef1?tdef1->getEnd(align1):get_ubound(a,sub1);
		SgExpression *ub2 = tdef2?tdef2->getEnd(align2):get_ubound(b,sub2);
		lb1 = calculateExp(lb1); ub1 = calculateExp(ub1);
		lb2 = calculateExp(lb2); ub2 = calculateExp(ub2);
		// compare the template size
		if (*lb1 != *lb2 || *ub1 != *ub2)
		{
if (!silentFlag)
printf("%s's Template start and end in dim %d is different with %s's dim %d.\n",a->identifier(),sub1+1,b->identifier(),sub2+1);
			return 0;
		}
		
		// compare dist code (block)
		if (ddef1->distCode(align1) != ddef2->distCode(align2))
		{
if (!silentFlag)
printf("%s's Distribute code in dim %d is different with %s's dim %d.\n",a->identifier(),sub1+1,b->identifier(),sub2+1);
			return 0;
		}
	}
	
if (!silentFlag)
printf("same data mapping pattern.\n");
	return 1;
}


// compare two shadow definitions
int compare_shadow(Shadow_def *sdef1, Shadow_def *sdef2)
{
	ASSERT(sdef1->getTag() == DEFINED && sdef2->getTag() == DEFINED);
	for (int i=0; i<MAX_SHADOW_DIM; i++)
	{
	  if (sdef1->getLower(i) != sdef2->getLower(i) ||
	      sdef1->getUpper(i) != sdef2->getUpper(i)) return 0;
	}
	return 1;
}

/***************************************************************************
		decide_optimize_major()
		column-major()
		row-major()
		to enhance locality
***************************************************************************/

// LHS have higher priority ,which have weight:4
// decide we use which major to optimize
int decide_optimize_major(SgStatement *s)
{
    SgForallStmt *forall = isSgForallStmt(s);
    ASSERT(forall);
    SgStatement *body = s->childList1(0);
    int i;

    forall_reference * lhsref = (forall_reference *) body->attributeValue(0, Forall_Ref);
    SgSymbol *lhsarraysymb = lhsref->getexpression()->symbol();

      // LHS have more priority
    int Major = getArrayStorage(lhsarraysymb);
    if (Major == COLUMN_MAJOR) Major = 4;
    else Major = -4;

    for (i=1; i<body->numberOfAttributes(Forall_Ref); i++) {
    	forall_reference * ref = (forall_reference *) body->attributeValue(i, Forall_Ref);
        SgSymbol *arraysymb = ref->getexpression()->symbol();
        int major = getArrayStorage(arraysymb);
	if (major == COLUMN_MAJOR) Major ++;
	else Major --;
    }
      // check the value of Major
    if (Major >= 0) return COLUMN_MAJOR;
    else return ROW_MAJOR;
}

// return 1, means optimization performed ; else 0 ,untouched
int column_major(SgStatement *s)
{
    SgForallStmt *forall = isSgForallStmt(s);
    ASSERT(forall);

    Forall_header_def * fhdef = (Forall_header_def *)s->attributeValue(0, FHDER_Def);
    SgStatement * body=s->childList1(0);
    forall_reference * leftref = (forall_reference *)body->attributeValue(0, Forall_Ref);
    SgArrayRefExp * larray = isSgArrayRefExp(leftref->getexpression());
    if (larray == NULL) return 0;

    int done_change=0;
    int change_in_loop=1;

    while (change_in_loop) 
    {
      change_in_loop=0;
      int lastdim = -1;
      for (int i=0; i<larray->numberOfSubscripts(); i++)
      {
	if (leftref->indexNumberOnDim(i) == -1) continue;
	if (lastdim == -1)
		lastdim=i; 
	else	// have to change two forall index
	if (leftref->indexNumberOnDim(i) < leftref->indexNumberOnDim(lastdim)) 
	{
		int dim1 = leftref->indexNumberOnDim(lastdim);
		int dim2 = leftref->indexNumberOnDim(i);
		// exchange two index
		SgExpression *index1 = forall->getIndex(dim1);
		SgExpression *index2 = forall->getIndex(dim2);
		SgExpression *tmp = index1->copyPtr();
		index1->replace(*index2->copyPtr());
		index2->replace(*tmp);
		// reset Index arrays.
		leftref->setIndex(lastdim,dim2,leftref->getIndex(i,dim2));
		leftref->setIndex(i,dim1,leftref->getIndex(lastdim,dim1));
		leftref->setIndex(lastdim,dim1,0);
		leftref->setIndex(i,dim2,0);
		done_change = change_in_loop = 1;
	}
	lastdim=i;
      }
    }

    if (done_change) {
	s->deleteAllAttribute(FHDER_Def);
  	body->deleteAllAttribute(Forall_Ref);
if (!silentFlag)
printf ("Column-Major optimize: change the index sequence in forall header.\n");
    }

    return done_change;
}

// return 1, means optimization performed ; else 0 ,untouched
int row_major(SgStatement *s)
{
    SgForallStmt *forall = isSgForallStmt(s);
    ASSERT(forall);

    Forall_header_def * fhdef = (Forall_header_def *)s->attributeValue(0, FHDER_Def);
    SgStatement * body=s->childList1(0);
    forall_reference * leftref = (forall_reference *)body->attributeValue(0, Forall_Ref);
    SgArrayRefExp * larray = isSgArrayRefExp(leftref->getexpression());
    if (larray == NULL) return 0;

    int done_change=0;
    int change_in_loop=1;

    while (change_in_loop) 
    {
      change_in_loop=0;
      int lastdim = -1;
      for (int i=larray->numberOfSubscripts()-1; i>=0; i--)
      {
	if (leftref->indexNumberOnDim(i) == -1) continue;
	if (lastdim == -1)
		lastdim=i; 
	else	// have to change two forall index
	if (leftref->indexNumberOnDim(i) < leftref->indexNumberOnDim(lastdim)) 
	{
		int dim1 = leftref->indexNumberOnDim(lastdim);
		int dim2 = leftref->indexNumberOnDim(i);
		// exchange two index
		SgExpression *index1 = forall->getIndex(dim1);
		SgExpression *index2 = forall->getIndex(dim2);
		SgExpression *tmp = index1->copyPtr();
		index1->replace(*index2->copyPtr());
		index2->replace(*tmp);
		// reset Index arrays.
		leftref->setIndex(lastdim,dim2,leftref->getIndex(i,dim2));
		leftref->setIndex(i,dim1,leftref->getIndex(lastdim,dim1));
		leftref->setIndex(lastdim,dim1,0);
		leftref->setIndex(i,dim2,0);
		done_change = change_in_loop = 1;
	}
	lastdim=i;
      }
    }

    if (done_change) {
	s->deleteAllAttribute(FHDER_Def);
  	body->deleteAllAttribute(Forall_Ref);
if (!silentFlag)
printf ("Row-Major optimize: change the index sequence in forall header.\n");
    }

    return done_change;
}

